import os.path, funcy
import biobox.image.volume as vol

from functools import partial

def get_all_biobox_paths(config):
    """
    Returns all paths listed in the biobox file
    """
    f = funcy.compose(
            partial(funcy.pluck, 'value'),
            funcy.flatten,
            partial(funcy.mapcat, funcy.itervalues))
    return f(config)


def remap_entries(xs):
    def remap(x):
        f = lambda i: vol.get_container_mount(i)['biobox_target']
        return funcy.update_in(x, ['value'], f)
    return list(map(remap, xs))


def remap_biobox_input_paths(args):
    return list(map(partial(funcy.walk_values, remap_entries), args))


def generate_biobox_file_content(args):
    import yaml
    output = {"version" : "0.9.0", "arguments" : args}
    return yaml.safe_dump(output, default_flow_style = False)


def create_biobox_directory(content):
    import tempfile
    dir_ = tempfile.mkdtemp()
    with open(os.path.join(dir_, "biobox.yaml"), "w") as f:
        f.write(content)
    return dir_



import funcy
import biobox.util as util
from requests.exceptions import ReadTimeout

def metadata_lookup(path, container_id):
    """
    Look up metadata about the docker container for the given path
    """
    return funcy.get_in(util.client().inspect_container(container_id), path)

def did_exit_succcessfully(id_):
    """
    Checks if the container exited with 0 exit code
    """
    return metadata_lookup(['State', 'ExitCode'], id_) == 0

def is_running(id_):
    """
    Checks if the container is running
    """
    return metadata_lookup(['State', 'Running'], id_)

def collect_runtime_metrics(container_id, interval = 15, warmup = 1):
    """
    Collects runtime metrics about the container at the specified intervals
    """
    import time
    time.sleep(warmup)

    stream = util.client().stats(container_id, decode = True, stream = True)

    stats = []
    while is_running(container_id):
        stats.append(next(stream))
        time.sleep(interval)
    return stats



import time, pytest
import helper           as hlp
import biobox.container as ctn
import biobox.util      as util

client = util.client()

def container(arg, wait = True, docker_args = {}):
    container = client.create_container(image='alpine:3.3', command='/bin/sleep {}'.format(str(arg)))
    id_ = container['Id']
    client.start(id_)
    if wait:
        client.wait(id_)
    return id_

def test_did_container_succesfully_exit():
    id_ = container(0.01)
    assert ctn.did_exit_succcessfully(id_)
    hlp.clean_up_container(id_)

def test_did_container_unssuccesfully_exit():
    id_ = container('bad-argument')
    assert not ctn.did_exit_succcessfully(id_)
    hlp.clean_up_container(id_)

def test_is_not_container_running():
    id_ = container('bad-argument')
    assert not ctn.is_running(id_)
    hlp.clean_up_container(id_)

def test_is_container_running():
    id_ = container(10, False)
    time.sleep(0.1)
    assert ctn.is_running(id_)
    client.kill(id_)
    hlp.clean_up_container(id_)

@pytest.mark.slow
def test_collect_metrics():
    id_ = container(4.5, False)
    stats = ctn.collect_runtime_metrics(id_, interval = 1)
    assert len(stats) >= 3
    assert 'read' in stats[0]
    hlp.clean_up_container(id_)



import tempfile, os.path, funcy
from functools import partial
import helper as hlp

import biobox.config       as cnf
import biobox.image.volume as vol

#################################################
#
# Test getting paths from biobox args
#
#################################################

def test_get_biobox_paths_with_single_file():
    path = hlp.create_tmp_file("file_0")
    biobox = [{"fastq": [
                        {"id": 0, "type": "paired", "value": path}
                        ]
                    }]
    assert cnf.get_all_biobox_paths(biobox) == [path]


def test_get_biobox_paths_with_two_files():
    path_0 = hlp.create_tmp_file("file_0")
    path_1 = hlp.create_tmp_file("file_1")
    biobox = [{"fastq": [
                        {"id": 0, "type": "paired", "value": path_0},
                        {"id": 0, "type": "paired", "value": path_1}
                        ]
                    }]
    assert cnf.get_all_biobox_paths(biobox) == [path_0, path_1]

#################################################
#
# Test remapping biobox input file paths
#
#################################################

def wrap(t, x):
    return [{t : x}]

def fastq(path):
    return {"id" : 0 , "value" : path, "type": "paired"}

def reference(path):
    return {"id" : 0 , "value" : path, "type": "reference"}

def test_remap_args_with_single_file_entry():
    host_file      = hlp.create_tmp_file("file_0")
    container_file = vol.get_container_mount(host_file)['biobox_target']
    assert cnf.remap_biobox_input_paths(wrap("fastq", [fastq(host_file)])) == \
            wrap("fastq", [fastq(container_file)])

def test_remap_args_with_single_directory_entry():
    host_dir      = hlp.create_tmp_dir("dir_0")
    container_dir = vol.get_container_mount(host_dir)['biobox_target']
    assert cnf.remap_biobox_input_paths(wrap("fasta_dir", [reference(host_dir)])) == \
            wrap("fasta_dir", [reference(container_dir)])



import docker
import docker.utils

def client(timeout = 60):
    args = docker.utils.kwargs_from_env(assert_hostname = False)
    args['version'] = '1.20'
    args['timeout'] = timeout
    return docker.Client(**args)



import docker.utils, funcy
import biobox.util         as util
import biobox.config       as cfg
import biobox.image.volume as vol

from functools import partial

def prepare_biobox_file(config):
    """
    Creates a biobox file in a temporary directory and returns a
    Docker volume string for that directory's location.
    """
    f = funcy.compose(
            vol.biobox_file,
            cfg.create_biobox_directory,
            cfg.generate_biobox_file_content,
            cfg.remap_biobox_input_paths)
    return f(config)


def prepare_volumes(config, output_directory):
    input_strings = vol.create_volume_string_set(cfg.get_all_biobox_paths(config))
    return input_strings + \
            [prepare_biobox_file(config)] + \
            [vol.output(output_directory)]

def create_container(image, config, output_directory, task = "default", docker_args = {}):
    volumes = prepare_volumes(config, output_directory)
    docker_args['volumes']     = list(map(vol.get_host_path, volumes))

    host_config = {'binds' : volumes}
    if 'mem_limit' in docker_args:
        host_config['mem_limit'] = docker_args['mem_limit']
        del docker_args['mem_limit']

    docker_args['host_config'] = util.client().create_host_config(**host_config)
    return util.client().create_container(image, task, **docker_args)



import pkg_resources, os, pkgutil
from setuptools import setup, find_packages
import biobox

def dependencies():
    file_ = pkg_resources.resource_filename(__name__, os.path.join('requirements', 'default.txt'))
    with open(file_, 'r') as f:
        return f.read().splitlines()


setup(
    name                 = 'biobox-py',
    version              = biobox.__version__,
    description          = 'Create and run biobox Docker containers',
    author               = 'bioboxes',
    author_email         = 'mail@bioboxes.org',
    url                  = 'http://bioboxes.org',
    install_requires     = dependencies(),

    packages             = find_packages(),
    include_package_data = True,

    classifiers = [
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Intended Audience :: Science/Research',
        'Operating System :: POSIX'
    ],
)



import biobox.util      as util
import biobox.exception
from functools import reduce

def get_image_tags(docker_dict):
    """
    Returns list of the names with and without tags for a given Docker image.
    E.g. returns ["python", "python:2.7"]
    """
    tags = docker_dict['RepoTags']
    if (tags is None) or (tags == [u'<none>:<none>']):
        return []
    else:
        return reduce(lambda acc, x: acc + [x, x.split(":")[0]], tags, [])

def get_image_digests(docker_dict):
    """
    Returns list all digests, if they exist, for a Docker image
    """
    digests = docker_dict['RepoDigests']
    if (digests is None) or (digests == [u'<none>@<none>']):
        return []
    else:
        return digests

def list_of_local_images():
    """
    Returns a set containing all local docker image tags and digests.
    """
    images = util.client().images()
    tags_and_digests = list(map(get_image_tags, images)) + list(map(get_image_digests, images))
    return set(reduce(lambda acc, x: acc + x, tags_and_digests, []))

def is_image_available_locally(name):
    """
    Does what you might expect.
    """
    return name in list_of_local_images()

def get_image(name):
    """
    Fetches the Docker image if it is not present locally.
    """
    if not is_image_available_locally(name):
        output = util.client().pull(name)
        if "error" in output:
            raise biobox.exception.NoImageFound(name)
    return True



import os, funcy
from functools import partial

BIOBOX_INPUT_MOUNT_LOC = "/bbx/mount"

def create_volume_string(host_dir, container_dir, read_only = True):
    """
    Creates the volume strings used by the Docker daemon for mounting volumes
    """
    access = "ro" if read_only else "rw"
    return ":".join([os.path.abspath(host_dir), container_dir, access])


def output(host_directory_path):
    """
    Creates volume string for the biobox output directory path
    """
    return create_volume_string(host_directory_path, "/bbx/output", False)


def input(host_directory_path):
    """
    Creates volume string for the biobox input directory path
    """
    return create_volume_string(host_directory_path, BIOBOX_INPUT_MOUNT_LOC, True)


def biobox_file(host_directory_path):
    """
    Creates volume string for the directory containing biobox.yaml
    """
    return create_volume_string(host_directory_path, "/bbx/input")


def host_directory(x):
    return os.path.dirname(os.path.abspath(x))


def get_host_path(volume_string):
    """
    Returns the host path from a Docker volume string
    """
    return volume_string.split(":")[0]


def get_container_path(host_directory, container_prefix_path = "/tmp"):
    """
    Returns a container directory location under the given prefix path.
    This is deterministic and always returns the same path for the given host path.
    """
    import hashlib
    from hashids import Hashids
    digest = funcy.rcompose(
            lambda x: hashlib.md5(x).hexdigest(),
            lambda x: int(x, base=16),
            Hashids(min_length=6).encode)
    return os.path.join(container_prefix_path, digest(host_directory))


def get_container_mount(path):
    """
    Returns dictionary with container mount locations for a given path.
    """
    host_dir = host_directory(path)
    cont_dir = get_container_path(host_dir, BIOBOX_INPUT_MOUNT_LOC)
    return {"host_dir"      : host_dir,
            "container_dir" : get_container_path(host_dir, BIOBOX_INPUT_MOUNT_LOC),
            "biobox_target" : os.path.join(cont_dir, os.path.basename(path)) }


def create_volume_string_set(paths):
    """
    Returns uniq list of volume strings for a given list host paths
    """
    f = funcy.rcompose(
            get_container_mount,
            lambda x: create_volume_string(x["host_dir"], x["container_dir"]))
    return funcy.distinct(map(f, paths))
